import { IERC20BridgeSamplerContract } from '@0x/contract-wrappers';
import { SignedOrder } from '@0x/types';
import { BigNumber } from '@0x/utils';

import { constants as marketOperationUtilConstants } from './constants';
import { DexSample, ERC20BridgeSource } from './types';

const { SOURCE_TO_ADDRESS } = marketOperationUtilConstants;

export class DexOrderSampler {
    private readonly _samplerContract: IERC20BridgeSamplerContract;

    /**
     * Generate sample amounts up to `maxFillAmount`.
     */
    public static getSampleAmounts(maxFillAmount: BigNumber, numSamples: number, expBase: number = 1): BigNumber[] {
        const distribution = [...Array<BigNumber>(numSamples)].map((v, i) => new BigNumber(expBase).pow(i));
        const stepSizes = distribution.map(d => d.div(BigNumber.sum(...distribution)));
        const amounts = stepSizes.map((s, i) => {
            return maxFillAmount
                .times(BigNumber.sum(...[0, ...stepSizes.slice(0, i + 1)]))
                .integerValue(BigNumber.ROUND_UP);
        });
        return amounts;
    }

    constructor(samplerContract: IERC20BridgeSamplerContract) {
        this._samplerContract = samplerContract;
    }

    public async getFillableAmountsAndSampleMarketBuyAsync(
        nativeOrders: SignedOrder[],
        sampleAmounts: BigNumber[],
        sources: ERC20BridgeSource[],
    ): Promise<[BigNumber[], DexSample[][]]> {
        const signatures = nativeOrders.map(o => o.signature);
        const [fillableAmount, rawSamples] = await this._samplerContract
            .queryOrdersAndSampleBuys(nativeOrders, signatures, sources.map(s => SOURCE_TO_ADDRESS[s]), sampleAmounts)
            .callAsync();
        const quotes = rawSamples.map((rawDexSamples, sourceIdx) => {
            const source = sources[sourceIdx];
            return rawDexSamples.map((sample, sampleIdx) => ({
                source,
                input: sampleAmounts[sampleIdx],
                output: sample,
            }));
        });
        return [fillableAmount, quotes];
    }

    public async getBatchFillableAmountsAndSampleMarketBuyAsync(
        nativeOrders: SignedOrder[][],
        sampleAmounts: BigNumber[][],
        sources: ERC20BridgeSource[],
    ): Promise<Array<[BigNumber[], DexSample[][]]>> {
        const signatures = nativeOrders.map(o => o.map(i => i.signature));
        const fillableAmountsAndSamples = await this._samplerContract
            .queryBatchOrdersAndSampleBuys(
                nativeOrders,
                signatures,
                sources.map(s => SOURCE_TO_ADDRESS[s]),
                sampleAmounts,
            )
            .callAsync();
        const batchFillableAmountsAndQuotes: Array<[BigNumber[], DexSample[][]]> = [];
        fillableAmountsAndSamples.forEach((sampleResult, i) => {
            const { tokenAmountsBySource, orderFillableAssetAmounts } = sampleResult;
            const quotes = tokenAmountsBySource.map((rawDexSamples, sourceIdx) => {
                const source = sources[sourceIdx];
                return rawDexSamples.map((sample, sampleIdx) => ({
                    source,
                    input: sampleAmounts[i][sampleIdx],
                    output: sample,
                }));
            });
            batchFillableAmountsAndQuotes.push([orderFillableAssetAmounts, quotes]);
        });
        return batchFillableAmountsAndQuotes;
    }

    public async getFillableAmountsAndSampleMarketSellAsync(
        nativeOrders: SignedOrder[],
        sampleAmounts: BigNumber[],
        sources: ERC20BridgeSource[],
    ): Promise<[BigNumber[], DexSample[][]]> {
        const signatures = nativeOrders.map(o => o.signature);
        const [fillableAmount, rawSamples] = await this._samplerContract
            .queryOrdersAndSampleSells(nativeOrders, signatures, sources.map(s => SOURCE_TO_ADDRESS[s]), sampleAmounts)
            .callAsync();
        const quotes = rawSamples.map((rawDexSamples, sourceIdx) => {
            const source = sources[sourceIdx];
            return rawDexSamples.map((sample, sampleIdx) => ({
                source,
                input: sampleAmounts[sampleIdx],
                output: sample,
            }));
        });
        return [fillableAmount, quotes];
    }
}
